/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "libvideocoding/mpeg2/syntax/demux.hh"
#include "libvideocoding/mpeg2/syntax/errors.hh"

namespace videocoding {

  MP2_PES_Demux::MP2_PES_Demux()
    : d_src(NULL)
  {
    for (int i=c_FirstStream;i<=c_LastStream;i++)
      {
	d_enabled[i-c_FirstStream]=false;
      }

    d_system_headers_enabled = false;
  }


  MP2_PES_Demux::~MP2_PES_Demux()
  {
    Reset();
  }


  void MP2_PES_Demux::SetPESSource(MP2_PES_Source& src)
  {
    d_src = &src;
    Reset();
  }


  void MP2_PES_Demux::Reset()
  {
    // Delete all packets that remained in the output queues.

    for (int i=c_FirstStream;i<=c_LastStream;i++)
      {
	while (!d_outputqueue[i-c_FirstStream].IsEmpty())
	  {
	    delete d_outputqueue[i-c_FirstStream].AskAndRemoveHead();
	  }
      }

    while (!d_miscqueue.IsEmpty())
      {
	delete d_miscqueue.AskAndRemoveHead();
      }
  }


  void MP2_PES_Demux::EnableChannel(MP2_StartCode StreamID,bool enable)
  {
    d_enabled[StreamID-c_FirstStream] = enable;

    if (!enable)
      {
	// Disabling channel. Delete all packet in its queue.

	while (!d_outputqueue[StreamID-c_FirstStream].IsEmpty())
	  {
	    delete d_outputqueue[StreamID-c_FirstStream].AskAndRemoveHead();
	  }
      }
  }


  void MP2_PES_Demux::EnableSystemHeaders(bool enable)
  {
    d_system_headers_enabled = enable;

    if (!enable)
      {
	while (!d_miscqueue.IsEmpty())
	  {
	    delete d_miscqueue.AskAndRemoveHead();
	  }
      }
  }


  bool MP2_PES_Demux::ReadAPacket()
  {
    MP2_SystemPacket* syspck = d_src->GetNextPacket();
    if (syspck==NULL)
      return false;

    MP2_PES_Packet* packet;
    if ((packet=dynamic_cast<MP2_PES_Packet*>(syspck)) != NULL)
      {
	int queue = packet->stream_id;

	if (queue<c_FirstStream || queue>c_LastStream)
	  {
	    delete packet;
	    throw Excpt_InvalidData("MP2 demultiplexer got PES packet with illegal stream id");
	  }
	else
	  {
	    if (d_enabled[queue-c_FirstStream])
	      { d_outputqueue[queue-c_FirstStream].Append(packet); }
	    else
	      { delete packet; }
	  }
      }
    else
      {
	d_miscqueue.Append(packet);
      }

    return true;
  }


  MP2_PES_Packet* MP2_PES_Demux::AskNextPacket(MP2_StartCode streamid)
  {
    int queueidx = streamid - c_FirstStream;

    Assert(d_enabled[queueidx]);


    // Read more packets until a packet of the requested type is received.

    while (d_outputqueue[queueidx].IsEmpty())
      {
	if (!ReadAPacket())
	  return NULL;
      }

    return d_outputqueue[queueidx].AskHead();
  }


  MP2_PES_Packet* MP2_PES_Demux::GetNextPacket(MP2_StartCode streamid)
  {
    MP2_PES_Packet* pck = AskNextPacket(streamid);
    if (pck) { d_outputqueue[streamid - c_FirstStream].RemoveHead(); }
    return pck;
  }


  bool MP2_PES_Demux::PacketPending(MP2_StartCode streamid) const
  {
    int queueidx = streamid - c_FirstStream;

    Assert(d_enabled[queueidx]);

    return !(d_outputqueue[queueidx].IsEmpty());
  }


  MP2_SystemPacket* MP2_PES_Demux::AskNextSystemPacket()
  {
    Assert(d_system_headers_enabled);

    // Read more packets until a packet of the requested type is received.

    while (d_miscqueue.IsEmpty())
      {
	if (!ReadAPacket())
	  return NULL;
      }

    return d_miscqueue.AskHead();
  }

  MP2_SystemPacket* MP2_PES_Demux::GetNextSystemPacket()
  {
    MP2_SystemPacket* pck = AskNextSystemPacket();
    if (pck) { d_miscqueue.RemoveHead(); }
    return pck;
  }

  bool MP2_PES_Demux::SystemPacketPending() const
  {
    Assert(d_system_headers_enabled);
    return !(d_miscqueue.IsEmpty());
  }





  MP2_PES_DemuxExtractor::MP2_PES_DemuxExtractor()
    : d_demux(NULL),
      d_initialized(false)
  {
  }


  MP2_SystemPacket* MP2_PES_DemuxExtractor::GetNextPacket()
  {
    Assert(d_demux != NULL);
    Assert(d_initialized);

    return d_demux->GetNextPacket(d_channel);
  }

  MP2_SystemPacket* MP2_PES_DemuxExtractor::AskNextPacket()
  {
    Assert(d_demux != NULL);
    Assert(d_initialized);

    return d_demux->AskNextPacket(d_channel);
  }

  bool MP2_PES_DemuxExtractor::PacketPending() const
  {
    Assert(d_demux != NULL);
    Assert(d_initialized);

    return d_demux->PacketPending(d_channel);
  }


  bool MP2_PES_DemuxExtractor::IsEOF() const
  {
    return const_cast<MP2_PES_DemuxExtractor*>(this)->AskNextPacket() != NULL;
  }

}
