/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "libvideocoding/mpeg2/syntax/video.hh"
#include "libvideocoding/mpeg2/syntax/system.hh"
#include "libvideocoding/mpeg2/syntax/errors.hh"
//#include <libvideogfx/error.hh>

#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <algorithm>


namespace videocoding {
  using namespace std;
  using namespace videogfx;

  static struct
  {
    uint16 framerate100;
    uint4 frame_rate_code;
  } frameratetable[] =
    {
      { 2500, MP2_FrameRate_25     },  // default
      { 2397, MP2_FrameRate_23_976 },
      { 2400, MP2_FrameRate_24     },
      { 2997, MP2_FrameRate_29_97  },
      { 3000, MP2_FrameRate_30     },
      { 5000, MP2_FrameRate_50     },
      { 5994, MP2_FrameRate_59_94  },
      { 6000, MP2_FrameRate_60     }
    };


  static void SkipToNextStartCode(BitReader& bs)
  {
    bs.SkipToByteBoundary();
    while (bs.PeekBits(24) != 0x000001 && !bs.IsEOF())
      bs.SkipBits(8);
  }


  static void CodeFrameRate(int framerate100,bool mayuseextension,
			    uint4& frame_rate_code, uint2& extnum, uint5& extden)
  {
    if (!mayuseextension)
      {
	extnum=0;
	extden=0;

	int bestidx=0;
	int err = abs(framerate100-frameratetable[0].framerate100);

	for (uint32 i=1;i<sizeof(frameratetable)/sizeof(frameratetable[0]);i++)
	  {
	    int newerr = abs(framerate100-frameratetable[i].framerate100);
	    if (newerr<err)
	      {
		err = newerr;
		bestidx = i;
	      }
	  }

	frame_rate_code = frameratetable[bestidx].frame_rate_code;

	if (err != 0)
	  {
	    char buf[100];
	    sprintf(buf,"unknown frame-rate %lf, coding at %lf.",
		    framerate100/100.0,
		    frameratetable[bestidx].framerate100/100.0);
	    MessageDisplay::Show(ErrSev_Warning,buf);
	  }
      }
    else
      {
	int best_err = abs(frameratetable[0].framerate100 - framerate100);
	int best_num=1, best_den=1, best_idx=0;

	for (extnum=1;extnum<=4;extnum++)
	  for (uint32 i=0;i<sizeof(frameratetable)/sizeof(frameratetable[0]);i++)
	    {
	      extden = frameratetable[i].framerate100*extnum/framerate100;

	      for (int dx=0;dx<=1;dx++)
		{
		  if (extden+dx >32 || extden+dx == 0)
		    continue;

		  int rate = frameratetable[i].framerate100*extnum/(extden+dx);
		  int err = abs(rate-framerate100);
		  if (err<best_err)
		    {
		      best_err = err;
		      best_num = extnum;
		      best_den = extden+dx;
		      best_idx = i;
		    }
		}
	    }


	frame_rate_code = frameratetable[best_idx].frame_rate_code;
	extnum          = best_num-1;
	extden          = best_den-1;

	if (best_err != 0)
	  {
	    int my_framerate = frameratetable[best_idx].framerate100 * best_num / best_den;
					    
	    char buf[100];
	    sprintf(buf,"cannot code at frame-rate %lf, using frame-rate %lf=%lf*%d/%d instead.",
		    framerate100/100.0,
		    my_framerate/100.0,
		    frameratetable[best_idx].framerate100/100.0,
		    best_num,best_den);
	    MessageDisplay::Show(ErrSev_Warning,buf);
	  }
      }
  }


  MP2_SequenceHeader::MP2_SequenceHeader(bool mpeg2)
  {
    SetDefaults(mpeg2);
  }


  void MP2_SequenceHeader::SetDefaults(bool mpeg2)
  {
    width = height = 0;
    bit_rate=0;
    vbv_buffer_size=0;
    
    aspect_ratio_information = MP2_AspectRatio_SAR_Square;
    framerate100 = 2500;
    constrained_parameters_flag = false;
    intra_matrix    = MP2_DefaultQuantMatrix_Intra;
    nonintra_matrix = MP2_DefaultQuantMatrix_Inter;
    
    IsMPEG2 = mpeg2;
    
    profile = MP2_Profile_Main;
    level   = MP2_Level_Main;
    progressive_sequence = true;
    chroma_format = MP2_Chroma_420;
    low_delay = false;
  }


  void MP2_SequenceHeader::Write(class BitBuffer& bs) const
  {
    // ===== SequenceHeader =====

    bs.AlignToByte0();
    bs.WriteBits(1,24);
    bs.WriteBits(MP2_StartCode_SequenceHeader,8);
    bs.WriteBits(width  & RightBits(12),12);
    bs.WriteBits(height & RightBits(12),12);
    bs.WriteBits(aspect_ratio_information,4);

    uint4 frame_rate_code;
    uint2 frame_rate_extension_n;
    uint5 frame_rate_extension_d;

    CodeFrameRate(framerate100,IsMPEG2,
		  frame_rate_code, frame_rate_extension_n, frame_rate_extension_d);

    bs.WriteBits(frame_rate_code,4);
    bs.WriteBits(bit_rate & RightBits(18),18);
    bs.WriteBits(1,1);
    bs.WriteBits(vbv_buffer_size & RightBits(10),10);
    bs.WriteBits(constrained_parameters_flag ? 1 : 0 ,1);

    if (HasNonDefaultIntraQMatrix())
      {
	const Scan* zigzag = AskScan(false);

	bs.WriteBits(1,1);
	for (int i=0;i<64;i++)
	  bs.WriteBits(intra_matrix.mat[zigzag[i].v][zigzag[i].u],8);
      }
    else
      bs.WriteBits(0,1);

    if (HasNonDefaultInterQMatrix())
      {
	const Scan* zigzag = AskScan(false);

	bs.WriteBits(1,1);
	for (int i=0;i<64;i++)
	  bs.WriteBits(nonintra_matrix.mat[zigzag[i].v][zigzag[i].u],8);
      }
    else
      bs.WriteBits(0,1);


    // ===== SequenceHeader-Extension =====

    if (IsMPEG2)
      {
	bs.AlignToByte0();
	bs.WriteBits(1,24);
	bs.WriteBits(MP2_StartCode_ExtensionStart,8);
	bs.WriteBits(MP2_ExtID_SequenceExtension , 4);
	bs.WriteBits((profile<<4) | level,8);
	bs.WriteBits(progressive_sequence ? 1 : 0 , 1);
	bs.WriteBits(chroma_format,2);
	bs.WriteBits((width >>12) & RightBits(2) , 2);
	bs.WriteBits((height>>12) & RightBits(2) , 2);
	bs.WriteBits((bit_rate>>18) & RightBits(12),12);
	bs.WriteBits(1,1);
	bs.WriteBits((vbv_buffer_size>>10) & RightBits(8),8);
	bs.WriteBits(low_delay ? 1 : 0,1);
	bs.WriteBits(frame_rate_extension_n,2);
	bs.WriteBits(frame_rate_extension_d,5);
      }
  }


  static const int frameratecode2framerate[16] =
    {
      0, 2400*1000/1001, 24*100,25*100, 3000*1000/1001, 30*100,50*100,6000*1000/1001,
      60*100, 0,0,0,0,0,0,0
    };


  void MP2_SequenceHeader::Read(class BitReader& bs)
  {
    /**** Read sequence header ****/

    if (bs.PeekBits(32) == (0x100 | MP2_StartCode_SequenceHeader))
      {
	bs.SkipBits(32);

	IsMPEG2 = false; /* This will be set to "true" if there is a Sequence-Extension following */
	width  = bs.GetBits(12);
	height = bs.GetBits(12);
	aspect_ratio_information = (MP2_AspectRatio)bs.GetBits(4);
	framerate100 = frameratecode2framerate[bs.GetBits(4)];

	/* TODO
	   if (seqh->Framerate == 0.0)
	   ShowNote(ErrSev_Warning,"MPEG stream error: framerate set to invalid value !");
	*/

	bit_rate = bs.GetBits(18);
	bs.SkipBits(1);
	vbv_buffer_size = bs.GetBits(10);
	constrained_parameters_flag = bs.GetBits(1);


	// Reset qmatrices or set them to new values.

	// Set Intra-Matrix
	if (bs.GetBits(1))
	  {
	    const Scan* zigzag = AskScan(false);

	    for (int i=0;i<64;i++)
	      intra_matrix.mat[zigzag[i].v][zigzag[i].u] =
		chroma_intra_matrix.mat[zigzag[i].v][zigzag[i].u] = bs.GetBits(8);
	  }
	else
	  {
	    intra_matrix = MP2_DefaultQuantMatrix_Intra;
	    chroma_intra_matrix = MP2_DefaultQuantMatrix_Intra;
	  }


	// Set Inter-Matrix
	if (bs.GetBits(1))
	  {
	    const Scan* zigzag = AskScan(false);

	    for (int i=0;i<64;i++)
	      nonintra_matrix.mat[zigzag[i].v][zigzag[i].u] =
		chroma_nonintra_matrix.mat[zigzag[i].v][zigzag[i].u] = bs.GetBits(8);
	  }
	else
	  {
	    nonintra_matrix = MP2_DefaultQuantMatrix_Inter;
	    chroma_nonintra_matrix = MP2_DefaultQuantMatrix_Inter;
	  }

	// fill MPEG-2 entries to match MPEG-1 semantics

	progressive_sequence = true;
	chroma_format = MP2_Chroma_420;
	low_delay = false;
	profile = MP2_Profile_MPEG1_Constrained;
	level   = MP2_Level_MPEG1_Constrained;

	SkipToNextStartCode(bs);
      }


    while (bs.PeekBits(32)==(0x100 | MP2_StartCode_ExtensionStart))
      {
	bs.SkipBits(32);

	uint4 extid = bs.GetBits(4);
	if (extid == MP2_ExtID_SequenceExtension)
	  {
	    /**** Decode Sequence-Extension ****/
    
	    IsMPEG2 = true;
    	    
	    profile = (MP2_ProfileID)bs.GetBits(4);
	    level   = (MP2_LevelID)bs.GetBits(4);
	    progressive_sequence = bs.GetBits(1);
	    chroma_format = (MP2_ChromaFormat)bs.GetBits(2);
	    width  |= bs.GetBits(2)<<12;
	    height |= bs.GetBits(2)<<12;
	    bit_rate |= bs.GetBits(12)<<18;
	    bs.SkipBits(1);
	    vbv_buffer_size |= bs.GetBits(8)<<10;
	    low_delay = bs.GetBits(1);
	    framerate100 *= bs.GetBits(2)+1;
	    framerate100 /= bs.GetBits(5)+1;
	  }
	else if (extid == MP2_ExtID_SequenceDisplayExtension)
	  {
	    bs.SkipBits(3);
	    if (bs.GetBits(1))
	      bs.SkipBits(3*8);
	    bs.SkipBits(29);

	    //MessageDisplay::Show(ErrSev_Warning,"unimplemented extension ignored: display extension");
	  }
	else if (extid == MP2_ExtID_QuantMatrixExtension)
	  {
	    const Scan* zigzag = AskScan(false);

	    if (bs.GetBits(1))
	      {
		for (int i=0;i<64;i++)
		  intra_matrix.mat[zigzag[i].v][zigzag[i].u] = bs.GetBits(8);
	      }

	    if (bs.GetBits(1))
	      {
		for (int i=0;i<64;i++)
		  nonintra_matrix.mat[zigzag[i].v][zigzag[i].u] = bs.GetBits(8);
	      }

	    if (bs.GetBits(1))
	      {
		for (int i=0;i<64;i++)
		  chroma_intra_matrix.mat[zigzag[i].v][zigzag[i].u] = bs.GetBits(8);
	      }

	    if (bs.GetBits(1))
	      {
		for (int i=0;i<64;i++)
		  chroma_nonintra_matrix.mat[zigzag[i].v][zigzag[i].u] = bs.GetBits(8);
	      }
	  }
	else if (extid == MP2_ExtID_SequenceScalableExtension)
	  {
	    MessageDisplay::Show(ErrSev_Warning,"unimplemented extension ignored: scalable extension");
	  }
	else
	  {
	    throw Excpt_InvalidData("unexpected extension after sequence header");
	  }

	SkipToNextStartCode(bs);
      }
  }


  ostream& operator<<(ostream& ostr,const MP2_SequenceHeader& sh)
  {
    ostr << "SequenceHeader:\n"
	 << "  MPEG-2:       " << (sh.IsMPEG2?"true":"false") << endl
	 << "  width:        " << sh.width << endl
	 << "  height:       " << sh.height << endl
	 << "  aspect code:  " << ((int)sh.aspect_ratio_information) << endl
	 << "  framerate:    " << sh.framerate100/100 << '.' << sh.framerate100%100 << endl
	 << "  bit rate:     " << sh.bit_rate*400 << " bits/s" << endl
	 << "  VBV buf.size: " << sh.vbv_buffer_size*16/8 << " kbyte" << endl
	 << "  constrained:  " << (sh.constrained_parameters_flag?"true":"false") << endl
      ;

    ostr << "  intra-qmatrix: ";
    if (sh.HasNonDefaultIntraQMatrix())
      ostr << endl << sh.intra_matrix;
    else
      ostr << "default\n";

    ostr << "  inter-qmatrix:";
    if (sh.HasNonDefaultInterQMatrix())
      ostr << endl << sh.nonintra_matrix;
    else
      ostr << "default\n";

    if (sh.IsMPEG2)
      {
	ostr << "  profile:      ";
	switch (sh.profile)
	  {
	  case MP2_Profile_Simple: ostr << "Simple\n"; break;
	  case MP2_Profile_Main:   ostr << "Main\n";   break;
	  case MP2_Profile_SNRScalable: ostr << "SNR\n"; break;
	  case MP2_Profile_SpatiallyScalable: ostr << "Spatial\n"; break;
	  case MP2_Profile_High:   ostr << "High\n"; break;
	  case MP2_Profile_MPEG1_Constrained: ostr << "MPEG-1-constrained\n"; break;
	  default: ostr << "unknown\n"; break;
	  }

	ostr << "  level:        ";
	switch (sh.level)
	  {
	  case MP2_Level_Low:    ostr << "Low\n"; break;
	  case MP2_Level_Main:   ostr << "Main\n"; break;
	  case MP2_Level_High1440: ostr << "High1440\n"; break;
	  case MP2_Level_High:   ostr << "High\n"; break;
	  case MP2_Level_MPEG1_Constrained: ostr << "MPEG-1-constrained\n"; break;
	  default: ostr << "unknown\n"; break;
	  }
	ostr << "  progr. seq.:  " << (sh.progressive_sequence?"true":"false") << endl;
	ostr << "  chroma format:";
	switch (sh.chroma_format)
	  {
	  case MP2_Chroma_420: ostr << "4:2:0\n"; break;
	  case MP2_Chroma_422: ostr << "4:2:2\n"; break;
	  case MP2_Chroma_444: ostr << "4:4:4\n"; break;
	  default: ostr << "unknown\n"; break;
	  }

	ostr << "  low delay:    " << (sh.low_delay?"true":"false") << endl;
      }

    return ostr;
  }


  bool MP2_SequenceHeader::HasNonDefaultIntraQMatrix() const
  {
    return intra_matrix != MP2_DefaultQuantMatrix_Intra;
  }


  bool MP2_SequenceHeader::HasNonDefaultInterQMatrix() const
  {
    return nonintra_matrix != MP2_DefaultQuantMatrix_Inter;
  }



  MP2_GOPHeader::MP2_GOPHeader()
  {
    time_code_drop_frame_flag = false;
    time_code_hours    = 0;
    time_code_minutes  = 0;
    time_code_seconds  = 0;
    time_code_pictures = 0;
    closed_gop  = false;
    broken_link = false;
  }


  void MP2_GOPHeader::AdvanceOnePicture(int fps)
  {
    if (++time_code_pictures == fps)
      {
	time_code_pictures=0;
	if (++time_code_seconds == 60)
	  {
	    time_code_seconds=0;
	    if (++time_code_minutes == 60)
	      {
		time_code_minutes=0;
		if (++time_code_hours == 24)
		  time_code_hours=0;
	      }
	  }
      }
  }


  void MP2_GOPHeader::Write(class BitBuffer& bs) const
  {
    bs.AlignToByte0();
    bs.WriteBits(1,24);
    bs.WriteBits(MP2_StartCode_GroupStart,8);
    bs.WriteBits(time_code_drop_frame_flag ? 1 : 0 ,1);
    bs.WriteBits(time_code_hours,5);
    bs.WriteBits(time_code_minutes,6);
    bs.WriteBits(1,1);
    bs.WriteBits(time_code_seconds,6);
    bs.WriteBits(time_code_pictures,6);
    bs.WriteBits(closed_gop  ? 1 : 0 ,1);
    bs.WriteBits(broken_link ? 1 : 0 ,1);
  }


  void MP2_GOPHeader::Read(class BitReader& bs)
  {
    Assert(bs.PeekBits(32)==(0x100 | MP2_StartCode_GroupStart));
    bs.SkipBits(32);

    time_code_drop_frame_flag = bs.GetBits(1);
    time_code_hours           = bs.GetBits(5);
    time_code_minutes         = bs.GetBits(6);
    bs.SkipBits(1);
    time_code_seconds         = bs.GetBits(6);
    time_code_pictures        = bs.GetBits(6);
    
    closed_gop  = bs.GetBits(1);
    broken_link = bs.GetBits(1);

    SkipToNextStartCode(bs);
  }


  ostream& operator<<(ostream& ostr,const MP2_GOPHeader& gop)
  {
    ostr << "GOP-Header:\n"
	 << "  Timecode:    " << ((int)gop.time_code_hours)
	 << ':' << setfill('0') << setw(2) << ((int)gop.time_code_minutes)
	 << ':' << setw(2) << ((int)gop.time_code_seconds)
	 << '.' << setw(2) << ((int)gop.time_code_pictures);
    if (gop.time_code_drop_frame_flag) ostr << " drop-frame-flag set";
    ostr << "\n  closed GOP:  " << (gop.closed_gop?"true":"false") << endl
	 << "  broken link: " << (gop.broken_link?"true":"false") << endl;

    return ostr;
  }


  MP2_PictureHeader::MP2_PictureHeader()
  {
    seqhdr = NULL;
    temporal_reference=0;
    picture_coding_type = MP2_PicType_I;
    vbv_delay = 0xFFFF;
    full_pel_forward_vector = false;
    full_pel_backward_vector = false;
    f_code[0][0]=f_code[0][1]=f_code[1][0]=f_code[1][1]=7;
    intra_dc_precision = 8;
    picture_structure=MP2_PicStruct_FramePicture;
    top_field_first = true;
    frame_pred_frame_dct = true;
    concealment_motion_vectors = false;
    q_scale_type=0;
    intra_vlc_format = 0;
    alternate_scan = false;
    
    has_coding_extension = false;
    
    repeat_first_field = false;
    progressive_frame = true;
    
    composite_display_flag = false;
    v_axis = 0;
    field_sequence = 0;
    sub_carrier = false;
    burst_amplitude = 0;
    sub_carrier_phase = 0;

    has_PTS=false;
    PTS=-1;
  }


  void MP2_PictureHeader::Read(class BitReader& bs)
  {
    if (bs.PeekBits(32)==(0x100 | MP2_StartCode_PictureStart))
      {
	bs.SkipBits(32);

	temporal_reference  = bs.GetBits(10);
	picture_coding_type = (MP2_PictureCodingType)bs.GetBits(3);
	vbv_delay           = bs.GetBits(16);
	
	full_pel_forward_vector  = false;
	full_pel_backward_vector = false;
	f_code[0][0] = 7;
	f_code[0][1] = 7;
	f_code[1][0] = 7;
	f_code[1][1] = 7;

	if (picture_coding_type == 2 ||
	    picture_coding_type == 3)
	  {
	    full_pel_forward_vector = bs.GetBits(1);
	    f_code[0][0] = f_code[0][1] = bs.GetBits(3);
	  }

	if (picture_coding_type == 3)
	  {
	    full_pel_backward_vector = bs.GetBits(1);
	    f_code[1][0] = f_code[1][1] = bs.GetBits(3);
	  }

	while (bs.GetBits(1)==1)
	  {
	    bs.SkipBits(8);
	  }

	// Set fields that are variable in MPEG-2 but constant in MPEG-1.
	// These will be overwritten again, when decoding MPEG-2 Picture Coding Extension.

	has_coding_extension = false;
	
	intra_dc_precision = 8;
	picture_structure  = MP2_PicStruct_FramePicture;
	concealment_motion_vectors = false;
	progressive_frame  = true;
	q_scale_type       = 0;
	intra_vlc_format   = 0;
	alternate_scan     = false;
	progressive_frame  = true;
	composite_display_flag = false;
	frame_pred_frame_dct = true;

	SkipToNextStartCode(bs);
      }


    while (bs.PeekBits(32)==(0x100 | MP2_StartCode_ExtensionStart))
      {
	bs.SkipBits(32);

	uint4 extid = bs.GetBits(4);
	if (extid == MP2_ExtID_PictureCodingExtension)
	  {
	    has_coding_extension = true;
	    
	    full_pel_forward_vector  = false;
	    full_pel_backward_vector = false;
	    f_code[0][0] = bs.GetBits(4);
	    f_code[0][1] = bs.GetBits(4);
	    f_code[1][0] = bs.GetBits(4);
	    f_code[1][1] = bs.GetBits(4);
	    
	    intra_dc_precision   = bs.GetBits(2)+8;
	    picture_structure    = (MP2_PictureStructure)bs.GetBits(2);
	    top_field_first      = bs.GetBits(1);
	    frame_pred_frame_dct = bs.GetBits(1);
	    concealment_motion_vectors = bs.GetBits(1);
	    q_scale_type         = bs.GetBits(1);
	    intra_vlc_format     = bs.GetBits(1);
	    alternate_scan       = bs.GetBits(1);
	    repeat_first_field   = bs.GetBits(1);
	    bs.SkipBits(1);
	    progressive_frame    = bs.GetBits(1);

	    composite_display_flag = bs.GetBits(1);
	    if (composite_display_flag)
	      {
		v_axis           = bs.GetBits(1);
		field_sequence   = bs.GetBits(3);
		sub_carrier      = bs.GetBits(1);
		burst_amplitude  = bs.GetBits(7);
		sub_carrier_phase= bs.GetBits(8);
	      }
	  }

	SkipToNextStartCode(bs);
      }
  }


  void MP2_PictureHeader::Write(class BitBuffer& bs) const
  {
    if (seqhdr)
      {
	Assert(has_coding_extension == seqhdr->IsMPEG2);
      }

    bs.AlignToByte0();
    bs.WriteBits(1,24);
    bs.WriteBits(MP2_StartCode_PictureStart,8);
    bs.WriteBits(temporal_reference,10);
    bs.WriteBits(picture_coding_type,3);
    bs.WriteBits(vbv_delay,16);

    if (picture_coding_type == MP2_PicType_P ||
	picture_coding_type == MP2_PicType_B)
      {
	if (!has_coding_extension) // if it is MPEG-1
	  {
	    Assert(f_code[0][0] == f_code[0][1]);

	    bs.WriteBits(full_pel_forward_vector ? 1 : 0 ,1);
	    bs.WriteBits(f_code[0][0],3);
	  }
	else
	  bs.WriteBits(7,4);
      }

    if (picture_coding_type == MP2_PicType_B)
      {
	if (!has_coding_extension) // if it is MPEG-1
	  {
	    Assert(f_code[1][0] == f_code[1][1]);
	    bs.WriteBits(full_pel_backward_vector ? 1 : 0 ,1);
	    bs.WriteBits(f_code[1][0],3);
	  }
	else
	  bs.WriteBits(7,4);
      }

    bs.WriteBits(0,1);


    // write picture coding extension if we are coding MPEG-2

    if (has_coding_extension)
      {
	bs.AlignToByte0();
	bs.WriteBits(1,24);
	bs.WriteBits(MP2_StartCode_ExtensionStart,8);
	bs.WriteBits(MP2_ExtID_PictureCodingExtension , 4);
	bs.WriteBits(f_code[0][0],4);
	bs.WriteBits(f_code[0][1],4);
	bs.WriteBits(f_code[1][0],4);
	bs.WriteBits(f_code[1][1],4);
	bs.WriteBits(intra_dc_precision-8,2);
	bs.WriteBits(picture_structure,2);
		     
	bs.WriteBits(top_field_first ? 1:0 ,1);
	bs.WriteBits(frame_pred_frame_dct ? 1:0 ,1);
	bs.WriteBits(concealment_motion_vectors ? 1:0 ,1);
	bs.WriteBits(q_scale_type,1);
	bs.WriteBits(intra_vlc_format,1);
	bs.WriteBits(alternate_scan?1:0 ,1);
	bs.WriteBits(repeat_first_field?1:0 ,1);

	if (seqhdr->chroma_format == MP2_Chroma_420)
	  bs.WriteBits(progressive_frame?1:0 ,1);
	else
	  bs.WriteBits(0,1);

	bs.WriteBits(progressive_frame?1:0 ,1);
	bs.WriteBits(composite_display_flag?1:0 ,1);

	if (composite_display_flag)
	  {
	    bs.WriteBits(v_axis,1);
	    bs.WriteBits(field_sequence,3);
	    bs.WriteBits(sub_carrier?1:0 ,1);
	    bs.WriteBits(burst_amplitude,7);
	    bs.WriteBits(sub_carrier_phase,8);
	  }
      }
    else
      {
	Assert(intra_dc_precision==8);
	Assert(picture_structure==MP2_PicStruct_FramePicture);
	Assert(frame_pred_frame_dct==true);
	Assert(q_scale_type==0);
	Assert(intra_vlc_format==0);
	Assert(alternate_scan==false);
      }
  }


  ostream& operator<<(ostream& ostr,const MP2_PictureHeader& pichdr)
  {
    ostr << "PictureHeader:\n"
	 << "  temporal reference:    " << pichdr.temporal_reference << endl
	 << "  picture_coding_type:   ";
    switch (pichdr.picture_coding_type)
      {
      case MP2_PicType_I: ostr << 'I'; break;
      case MP2_PicType_P: ostr << 'P'; break;
      case MP2_PicType_B: ostr << 'B'; break;
      case MP2_PicType_D: ostr << 'D'; break;
      default: ostr << "UNKNOWN !"; break;
      }

    ostr << "\n  vbv_delay:             " << pichdr.vbv_delay << endl
	 << "  fullpel forward mv:    " << (pichdr.full_pel_forward_vector?"true":"false") << endl
	 << "  fullpel backward mv:   " << (pichdr.full_pel_backward_vector?"true":"false") << endl;
    for (int i=0;i<2;i++)
      for (int j=0;j<2;j++)
	{
	  ostr << "  fcode[" << i << "][" << j << "] ("
	       << (i?"bw/":"fw/") << (j?'v':'h')
	       << "):    " << ((int)pichdr.f_code[i][j]) << endl;
	}

    if (pichdr.has_coding_extension)
      {
	ostr << "  intra dc precision:    " << ((int)pichdr.intra_dc_precision) << endl
	     << "  picture structure:     ";
	switch (pichdr.picture_structure)
	  {
	  case MP2_PicStruct_TopField: ostr << "top field\n"; break;
	  case MP2_PicStruct_BottomField: ostr << "bottom field\n"; break;
	  case MP2_PicStruct_FramePicture: ostr << "frame picture\n"; break;
	  }
 
	ostr << "  top field first:       " << (pichdr.top_field_first?"true":"false") << endl
	     << "  frame pred frame dct:  " << (pichdr.frame_pred_frame_dct?"true":"false") << endl
	     << "  concealment mvs:       " << (pichdr.concealment_motion_vectors?"true":"false") << endl
	     << "  q-scale type:          " << ((int)pichdr.q_scale_type) << endl
	     << "  intra vlc format:      " << ((int)pichdr.intra_vlc_format) << endl
	     << "  alternate scan:        " << (pichdr.alternate_scan?"true":"false") << endl
	     << "  repeat first field:    " << (pichdr.repeat_first_field?"true":"false") << endl
	     << "  progressive frame:     " << (pichdr.progressive_frame?"true":"false") << endl;
          
	if (pichdr.composite_display_flag)
	  {
	    ostr << "  v-axis:                " << ((int)pichdr.v_axis) << endl
		 << "  field sequence:        " << ((int)pichdr.field_sequence) << endl
		 << "  sub carrier:           " << ((int)pichdr.sub_carrier) << endl
		 << "  burst amplitude:       " << ((int)pichdr.burst_amplitude) << endl
		 << "  sub carrier phase:     " << ((int)pichdr.sub_carrier_phase) << endl;
	  }
      }

    ostr << "  PTS:                   ";
    if (pichdr.has_PTS) ostr << ((int32)pichdr.PTS) << endl;
    else ostr << "none\n";

    return ostr;
  }


  void MP2_SliceHeader::Read(class BitReader& bs)
  {
    Assert(pichdr);

    uint32 sc = bs.PeekBits(32);
    MP2_StartCode startcode = (MP2_StartCode)(sc & 0xFF);

    Assert((sc & 0xFFFF00) == 0x100);
    Assert(IsMPEG_SLICE_START(startcode));

    bs.SkipBits(24);

    slice_vertical_position = bs.GetBits(8)-1;

    if (pichdr->seqhdr->height > 2800)
      slice_vertical_position += bs.GetBits(3)<<7;

    /*
      if (d_scalability_mode == DataPartitioning)
      slicehdr.priority_breakpoint = bs.GetBits(7);
    */

    qcode = bs.GetBits(5);
    if (bs.GetBits(1)==1)
      {
	intra_slice = bs.GetBits(1);
	bs.SkipBits(7);

	while (bs.GetBits(1)==1)
	  bs.SkipBits(8);
      }
    else
      intra_slice=false;

    // bs.SkipToByteBoundary();
  }


  ostream& operator<<(ostream& ostr,const MP2_SliceHeader& slicehdr)
  {
    ostr << "SliceHeader:\n"
	 << "  vertical pos:         " << slicehdr.slice_vertical_position << endl
	 << "  quantiser scale code: " << ((int)slicehdr.qcode) << endl
	 << "  intra slice:          " << (slicehdr.intra_slice?"true":"false") << endl;

    return ostr;
  }


  void MP2_SliceHeader::Write(BitBuffer& bs) const
  {
    bs.AlignToByte0();
    bs.WriteBits(0x000001,24);

    if (pichdr->seqhdr->height<=2800)
      {
	bs.WriteBits(slice_vertical_position+1,8);
      }
    else
      {
	uint3 low7bits = slice_vertical_position & RightBits(7);
	bs.WriteBits(low7bits,8);
	bs.WriteBits((slice_vertical_position>>7) & RightBits(3),3);
      }

    // TODO: jetzt kommen eigentlich die Scalability-Felder

    bs.WriteBits(qcode,5);
    if (CodeIntraSlice)
      {
	bs.WriteBits(1,1);
	bs.WriteBits(intra_slice ? 1:0 ,1);
	bs.WriteBits(0,7);
      }
  
    bs.WriteBits(0,1);
  }




  static Scan scan_zigzag[64] =
    { { 0, 0 },  { 1, 0 },  { 0, 1 },  { 0, 2 },
      { 1, 1 },  { 2, 0 },  { 3, 0 },  { 2, 1 },
      { 1, 2 },  { 0, 3 },  { 0, 4 },  { 1, 3 },
      { 2, 2 },  { 3, 1 },  { 4, 0 },  { 5, 0 },
      { 4, 1 },  { 3, 2 },  { 2, 3 },  { 1, 4 },
      { 0, 5 },  { 0, 6 },  { 1, 5 },  { 2, 4 },
      { 3, 3 },  { 4, 2 },  { 5, 1 },  { 6, 0 },
      { 7, 0 },  { 6, 1 },  { 5, 2 },  { 4, 3 },
      { 3, 4 },  { 2, 5 },  { 1, 6 },  { 0, 7 },
      { 1, 7 },  { 2, 6 },  { 3, 5 },  { 4, 4 },
      { 5, 3 },  { 6, 2 },  { 7, 1 },  { 7, 2 },
      { 6, 3 },  { 5, 4 },  { 4, 5 },  { 3, 6 },
      { 2, 7 },  { 3, 7 },  { 4, 6 },  { 5, 5 },
      { 6, 4 },  { 7, 3 },  { 7, 4 },  { 6, 5 },
      { 5, 6 },  { 4, 7 },  { 5, 7 },  { 6, 6 },
      { 7, 5 },  { 7, 6 },  { 6, 7 },  { 7, 7 } };

  static Scan scan_alternate[64] = 
    { { 0, 0 },  { 0, 1 },  { 0, 2 },  { 0, 3 },
      { 1, 0 },  { 1, 1 },  { 2, 0 },  { 2, 1 },
      { 1, 2 },  { 1, 3 },  { 0, 4 },  { 0, 5 },
      { 0, 6 },  { 0, 7 },  { 1, 7 },  { 1, 6 },
      { 1, 5 },  { 1, 4 },  { 2, 3 },  { 2, 2 },
      { 3, 0 },  { 3, 1 },  { 4, 0 },  { 4, 1 },
      { 3, 2 },  { 3, 3 },  { 2, 4 },  { 2, 5 },
      { 2, 6 },  { 2, 7 },  { 3, 4 },  { 3, 5 },
      { 3, 6 },  { 3, 7 },  { 4, 2 },  { 4, 3 },
      { 5, 0 },  { 5, 1 },  { 6, 0 },  { 6, 1 },
      { 5, 2 },  { 5, 3 },  { 4, 4 },  { 4, 5 },
      { 4, 6 },  { 4, 7 },  { 5, 4 },  { 5, 5 },
      { 5, 6 },  { 5, 7 },  { 6, 2 },  { 6, 3 },
      { 7, 0 },  { 7, 1 },  { 7, 2 },  { 7, 3 },
      { 6, 4 },  { 6, 5 },  { 6, 6 },  { 6, 7 },
      { 7, 4 },  { 7, 5 },  { 7, 6 },  { 7, 7 } };

  const Scan* AskScan(bool alternate_scan)
  {
    if (alternate_scan) return scan_alternate;
    else return scan_zigzag;
  }


  static const int qcode2qscale[2][32] =
    {
      //  - 1 2 3 4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29  30  31
      { 0,2,4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56,58, 60, 62 },
      { 0,1,2,3,4, 5, 6, 7, 8,10,12,14,16,18,20,22,24,28,32,36,40,44,48,52,56,64,72,80,88,96,104,112 }
    };

  const int* AskQScaleTab(bool nonuniform)
  {
    if (nonuniform) return qcode2qscale[1];
    else            return qcode2qscale[0];
  }


  QMatrix_8x8 MP2_DefaultQuantMatrix_Intra =
    {
      {
	{  8, 16, 19, 22, 26, 27, 29, 34 },
	{ 16, 16, 22, 24, 27, 29, 34, 37 },
	{ 19, 22, 26, 27, 29, 34, 34, 38 },
	{ 22, 22, 26, 27, 29, 34, 37, 40 },
	{ 22, 26, 27, 29, 32, 35, 40, 48 },
	{ 26, 27, 29, 32, 35, 40, 48, 58 },
	{ 26, 27, 29, 34, 38, 46, 56, 69 },
	{ 27, 29, 35, 38, 46, 56, 69, 83 }
      }
    };

  QMatrix_8x8 MP2_DefaultQuantMatrix_Inter =
    {
      {
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 },
	{ 16, 16, 16, 16, 16, 16, 16, 16 }
      }
    };

}
